'
' -----[ Title ]------------------------------------------------------
'
' File......: SNAP.BAS
' Purpose...: Prototype of a S.N.A.P controlled I/O device
' Author....: Mike Eitel
' Version...: 1.10AVR
' MCU.......: AVR 8515 or 8535
' Started...: 991002
' Updated...: 991024
'
' -----[ Program Description ]----------------------------------------
'
' This program implements the S.N.A.P protocol in BASCOM for an universal I/O device with stepper driver
' It uses STK200 development kit for testing purpouse.
' It's based on 16-bit CRC-CCITT as error detection method to give secure data transfer.
'
' This File is partialy based on a program from Christer Johansson HTH and
' I do not intent to touch any of his rights !!!
' So for correct use of his copyrights please inform as marked below!
' For more information about S.N.A.P see URL http://www.hth.com/snap/
'
' -----[ Disclaimer ]-------------------------------------------------
'
' This example is offered on an "AS IS" basis, no warranty expressed or implied.
' The programers disclaim liability of any damages associated with the use of the hardware or software
' described herein. You use it on your own risk. We are not able to provide any free support.
'
' Copyright (c) 1998-1999 HTH on theirs parts all rights reserved
' Copyright (c) 1999 Mike Eitel requests on his changed parts all rights reserved
'
' -----[ Revision History ]-------------------------------------------
'
' 980529 - Version 1.00    Initial release                                       Christer Johansson HTH
' 980919 - Version 1.01    Revised for S.N.A.P Version 0.91                      Christer Johansson HTH
' 990130 - Version 1.02    Revised for S.N.A.P Version 1.00                      Christer Johansson HTH
' 991002 - Version 1.00AVR S.N.A.P Ver 1.00 for AVR 8515                         Mike Eitel
' 991011 - Version 1.10AVR S.N.A.P Ver 1.10 with stepper commands for AVR 8515   Mike Eitel
'
' -----[ Aliases ]----------------------------------------------------
'
Stepperport Alias Porta                                       ' Status of the Stepper bridge Transistors
Commandport Alias Portb                                       ' Commands pins
Ledport Alias Portc                                           ' LED pins showing values
'
' -----[ Constants ]--------------------------------------------------
'
' The packet structure is defined in the received packets first two bytes (HDB2 and HDB1).
' Overview of header definition bytes (HDB2 and HDB1)
'
'                            HDB2              HDB1
'                     +-----------------+-----------------+
'                     | D D S S P P A A | C E E E N N N N |
'                     +-----------------+-----------------+
'
' The following packet structure is used.
'
'                      DD=01     - 1 Byte destination address
'                      SS=01     - 1 Byte source address
'                      PP=00     - No protocol specific flags
'                      AA=01     - Acknowledge is required
'                                         C=0/1     - Command mode not / is supported
'                                         EEE=100   - 16-bit CRC-CCITT
'                                         NNNN=0101 - 5 Byte data
'
Const Preamble = &B01010011                                   ' Preamble byte
Const Sync = &B01010100                                       ' Synchronisation byte
Const Crcpoly = &H1021                                        ' CRC-CCITT
Const Chdb2 = &B01010001                                      ' &B01010001 DD=01 - 1 Byte destination address
'                                                                          SS=01 - 1 Byte source address
'                                                                          PP=00 - No protocol specific flags
'                                                                          AA=01 - Acknowledge is required
Const Anshdb2 = &B01010000                                    ' No Ack request 01 01 00 00
Const Ackhdb2 = &B01010010                                    ' Short Ack      01 01 00 10
Const Nakhdb2 = &B01010011                                    ' Short Nak      01 01 00 11


Const Ackhdb1 = &B01000001                                    ' Short Ack  0 100 0001
Const Chdb1r = &B01000101                                     ' &B01000101 C =   1/0  - Command mode / not supported
Const Chdb1w = &B11000101                                     '            Eee = 100  - 16-bit Crc -ccitt
'                                                                          NNNN= 0101 - 5 Byte data


Const Myaddress = &H04                                        ' Address for this node (1-255)

Const Cmaxchar = 26                                           ' Number of characters for serial buffer
Const Numofdatabytes = 5                                      ' Number of transmitted Databytes

Const Dataarea = 16                                           ' Size of the Databuffer
Const Dataareanr = 4                                          ' Size of the Databuffer divided by 4
Const Paramarea = 16                                          ' Size of the Parameters memory
Const Paramareanr = 4                                         ' Size of the Parameters memory divided by 4


Const Timer0_reload = &H80                                    ' Base constant for main timing
Const Timer1_reload = &H1FFF                                  ' Default constant for scanned sending

' -----[ Variables ]--------------------------------------------------
'
Dim Temp1 As Byte                                             ' Temporary Variable
Dim Temp2 As Byte                                             ' Temporary Variable
Dim Temp_crc As Word                                          ' Temporary Variable
Dim Crc As Word                                               ' CRC word variable
Dim Crc2 As Byte                                              ' CRC hi byte variable
Dim Crc1 As Byte                                              ' CRC lo byte variable
Dim Hdb2 As Byte                                              ' Header Definition Byte 2
Dim Hdb1 As Byte                                              ' Header Definition Byte 1
Dim Sab1 As Byte                                              ' What node sent this packet
Dim Dab1 As Byte                                              ' What node should have this paket
Dim Db(numofdatabytes) As Byte                                ' Packet receivedata ( 2-5 ) with command (1)
Dim Sdb(numofdatabytes) As Byte                               ' Packet senddata ( 2-5 ) with command (1)
Dim Hisaddress As Byte                                        ' Byte containing the last requester adress
Dim Onebyte As Byte                                           ' The received datas
'                            Commands
'                         +-----------------+                   W    = 0/1 Read / Write
'                         | W C C C A A A A |                   CCC  = 0. 7  Different command categories
'                         +-----------------+                        = 7 prioritised commands
'                                                                    = 6
'                                                                    = 5
'                                                                    = 4
'                                                                    = 3
'                                                                    = 2 EPROM Data Area
'                                                                    = 1 SRAM Data Area
'                                                                    = 0
'                                                               AAAA = 0..15 Pointer for values
'
Dim Commands As Byte                                          ' Byte to recognize the requested commands
Dim Commandcat As Byte                                        ' Byte with the command categories
Dim Commandnr As Byte                                         ' Byte with the command pointer
Dim Commandmodus As Bit                                       ' Bit that commands are used
Dim Commandsend As Bit                                        ' Bit that 1= get command / 0= AVR has to resend data
Dim Commandprios As Bit                                       ' Bit to recognize prioritised commands

Dim I As Byte                                                 ' Help byte for data transfer
Dim J As Byte                                                 ' Help byte for data transfer
Dim K As Byte                                                 ' Help byte for data transfer

Dim Timer1_reload2 As Integer                                 ' For adjusted scanned sending

Dim Bitsreceived As Byte                                      ' Byte for signalling several received characters
Dim Bitbuffer(cmaxchar) As Byte                               ' Receive serial-buffer
Dim Bitcounter As Byte                                        ' Byte counter of received characters
Dim Bitreader As Byte                                         ' Byte counter of regarded characters

Dim Countmeup As Word
Dim Stepstatus As Byte                                        ' Byte containing the value for the Bridge transistors
Dim Laststepstatus As Byte                                    ' Byte containing the last value for the Bridge
Dim Tempstepstatus As Byte                                    ' Help Byte for the Bridge transistors

Dim Datas(dataarea) As Sram Byte                              ' At &H100       ' Place to keep the dynamic datas
Dim Parameters(paramarea) As Byte                             ' Eram Byte  AT xxx            ' Place to keep the dynamic datas


' -----[ Initialization ]---------------------------------------------
'
' Define serial communication parameters
_main:
   $baud = 9600
   $crystal = 4000000

   Config Porta = Output                                      ' Port for controlling the stepper phase transistors
   Config Portb = Output                                      '
   Config Portc = Output                                      '
   Config Timer0 = Timer , Prescale = 64                      ' Main timer
   Config Timer1 = Timer , Prescale = 256                     '

   Porta = &HFF                                               ' define startup status of Port A
   Portb = &HFF                                               ' define startup status of Port B
   Portc = &HFF                                               ' define startup status of Port C
   Tccr1a = &B00010011                                        ' *** dammned trick to get the OC1B output blinking

   Timer0 = Timer0_reload                                     ' used for motor frequency adjustment
   Timer1 = Timer1_reload                                     '

   Bitsreceived = 0                                           ' Set to defined Startup value
   Bitcounter = 0                                             ' Set to defined Startup value
   Bitreader = 0                                              ' Set to defined Startup value
   For I = 1 To Cmaxchar
       Bitbuffer(i) = 0                                       ' Initialize the buffer with 0 against garbage
   Next I

   Commands = 0                                               ' Set to defined Startup value
   Stepstatus = &H0F                                          ' Set to defined Startup value
   Laststepstatus = &H0F                                      ' Set to defined Startup value
   Countmeup = 0                                              ' Set to defined Startup value


   'define ISR when serial input
   On Urxc Rec_isr                                            'Define serial receive ISR

   'define ISR when timer 0 overflows
   On Timer0 Timer0_isr                                       ' Base jump of the system. This timer has the highest
                                                              ' "priority" cause it is important to have a constant
                                                              ' stepper frequency to avoid speed oscillations
                                                              ' define ISR when timer 1 overflows
   On Timer1 Timer1_isr                                       '?

   ' Setup interrupts to defined startup
   Enable Timer0                                              ' Enable base timer of the system.
   Disable Timer1                                             ' Enable base timer of the system.
   Enable Urxc                                                ' Enable receive isr
   Disable Utxc                                               ' Disable send isr
   Disable Udre                                               ' Disable empty buffer isr

'------[ Program ]----------------------------------------------------
'
Print Chr(preamble) ; Chr(preamble) ; Myaddress ; Chr(preamble) ; Chr(preamble)

Enable Serial                                                 ' Allow protokoll treatment starts
Enable Interrupts                                             ' Start the ISR system from this line on



_com_1:                                                       ' Startingpoint for telegramms
   If Bitsreceived >= 1 Then                                  ' Received something
      Bitreader = Bitcounter - Bitsreceived
      Bitreader = Bitreader + 1
      Temp1 = Bitbuffer(bitreader)
   Else
      Goto _com_1
   End If
   Bitsreceived = Bitsreceived - 1

   ' If received data is SYNC byte read next eleven bytes from master, if not return to start
   If Temp1 <> Sync Then
                                      'Print "WRONG " ; Chr(sync) ; "-" ; Bitcounter ; "-" ; Bitreader ; "->" ; Temp1 ; "-" ; Bitbuffer(1)
      Goto _com_1
   Else
     Incr Bitreader
     If Bitreader >= Cmaxchar Then                            ' Buffer full
        Bitreader = 1                                         ' Rewrap character serial counter
   End If

_com_2:

   If Bitsreceived < 11 Then
      Goto _com_2
   End If

   ' Get telegramm in binary mode
   Hdb2 = Bitbuffer(bitreader)                                ' HDB2
   Incr Bitreader
   Hdb1 = Bitbuffer(bitreader)                                ' HDB1
   Incr Bitreader
   Dab1 = Bitbuffer(bitreader)                                ' DAB1
   Incr Bitreader
   Sab1 = Bitbuffer(bitreader)                                ' Sab1
   Incr Bitreader
   Db(5) = Bitbuffer(bitreader)                               ' DB5
   Incr Bitreader
   Db(4) = Bitbuffer(bitreader)                               ' DB4
   Incr Bitreader
   Db(3) = Bitbuffer(bitreader)                               ' DB3
   Incr Bitreader
   Db(2) = Bitbuffer(bitreader)                               ' DB2
   Incr Bitreader
   Db(1) = Bitbuffer(bitreader)                               ' DB1 / Command
   Incr Bitreader
   Crc2 = Bitbuffer(bitreader)                                ' CRC2
   Incr Bitreader
   Crc1 = Bitbuffer(bitreader)                                ' CRC1
   Incr Bitreader

   Bitcounter = 1
   Bitsreceived = 0

   ' Packet header check routine
   ' Check HDB2 to see if MCU is capable to use the packet structure, if not goto Startingpoint
   If Hdb2 <> Chdb2 Then
      Goto _com_1
   End If

   ' Check HDB1 to see if MCU is capable to use the packet structure, if not goto Startingpoint
   Select Case Hdb1
          Case Chdb1w :
                        Commandmodus = 1                      ' Check for command modus
                        Commands = Db(1)
          Case Chdb1r :
                        Commandmodus = 0                      ' Check for pure data byte receiving modus
          Case Else :
          Goto _com_1
   End Select

   ' Address check routine checks if this is the node addressed, if not goto Startingpoint
   If Dab1 <> Myaddress Then
      Goto _com_1
   End If

   Hisaddress = Sab1                                          'keep in mind his address

   ' CRC check routine and calculate CRC for all received bytes
   Crc = 0
   Temp2 = Hdb2
   Gosub Calc_crc
   Temp2 = Hdb1
   Gosub Calc_crc
   Temp2 = Dab1
   Gosub Calc_crc
   Temp2 = Sab1
   Gosub Calc_crc
   Temp2 = Db(5)
   Gosub Calc_crc
   Temp2 = Db(4)
   Gosub Calc_crc
   Temp2 = Db(3)
   Gosub Calc_crc
   Temp2 = Db(2)
   Gosub Calc_crc
   Temp2 = Db(1)
   Gosub Calc_crc
   Temp2 = Crc2
   Gosub Calc_crc
   Temp2 = Crc1
   Gosub Calc_crc

   ' Check if there was any CRC errors, if so send NAK
   If Crc <> 0 Then
      Gosub _nak                                              ' Gosub _nak Routine
      Goto _com_1                                             ' Restart reading process
   End If

   ' No CRC errors in packet so check what to do and then start writing data or do according commands
   If Commandmodus = 0 Then
      Ledport = Onebyte Xor &HFF
      Onebyte = Db(1)
      Gosub _ack
   Else
      If Commands > &H80 Then                                 ' decide if send or re.read action requested
         Commandsend = 1
      Else
         Commandsend = 0
      End If
      Commandcat = Commands And &H70                          ' Decide command categorie
      Commandcat = Commandcat / &H0F                          ' between  0 ..7
      Commandnr = Commands And &H0F                           ' Decide command pointer

      Commandport = Commands Xor &HFF

      Select Case Commandcat                                  ' Decide what kind of command categorie to treat
             Case 1 :                                         ' Normal writing and re-reading SRAM
                      If Commandnr > Dataareanr Then          ' protection against area size error
                         Goto _no_command
                      Else
                         Gosub _ack
                      End If
                      If Commandsend = 1 Then
                         For I = 1 To 4 Step 1                ' 1..4
                             J = I + 1                        ' 2..5
                             K = Commandnr * 4                ' 0    4    8   ...
                             K = K + I                        ' 1..4 5..8 ...
                             Datas(k) = Db(j)                 ' writing
                         Next I
                      Else
                         Sdb (1) = Commandcat * 16
                         Sdb (1) = Sdb(1) + Commandnr
                         For I = 1 To 4 Step 1                ' 1..4
                             J = I + 1                        ' 2..5
                             K = Commandnr * 4                ' 0     4     8 ...
                             K = K + I                        ' 1..4  5..8  ....
                             Sdb(j) = Datas (k)               ' reading
                         Next I
                         Gosub _send
                      End If
             Case 2                                           ' Writing and re-reading EPROM
                      If Commandnr > Paramareanr Then         ' protection against area size error
                         Goto _no_command
                      Else
                         Gosub _ack
                      End If
                      If Commandsend = 1 Then
                         For I = 1 To 4 Step 1                ' 1..4
                             J = I + 1                        ' 2..5
                             K = Commandnr * 4                ' 0    4    8   ...
                             K = K + I                        ' 1..4 5..8 ...
                             Parameters(k) = Db(j)            ' writing
                         Next Iii
                      Else
                         Sdb (1) = Commandcat * 16
                         Sdb (1) = Sdb(1) + Commandnr
                         For I = 1 To 4 Step 1                ' 1..4
                             J = I + 1                        ' 2..5
                             K = Commandnr * 4                ' 0     4     8 ...
                             K = K + I                        ' 1..4  5..8  ....
                             Sdb(j) = Parameters (k)          ' reading
                         Next I
                         Gosub _send
                      End If
             Case 7 :                                         ' Spezial / priorised commands
                      Select Case Commandnr
                             'Case 0 :
                                   '
                             Case 14 :
                                        Disable Timer1        ' Disallow cyclic send
                                        Gosub _ack
                             Case 15 :
                                        Enable Timer1         ' Disallow cyclic send
                                        Gosub _ack
                             Case Else :
                                        Goto _no_command      ' NOT recognized command
                             End Select
             Case Else :
_no_command:
                             Gosub _nak                       ' NOT recognized command
      End Select                                              ' Of command modus decision tree
   End If                                                     ' Of modus seperation

' Done, go back to Start and wait for a new packet
Goto _com_1
   End If                                                     'of treating a correct telegramm
Goto _com_1                                                   ' wait for the next telegramm


' End of the main program
' -----[ End ]--------------------------------------------------------



' -----[ Subroutines ]------------------------------------------------
'

' Subroutine for calculating CRC value in variable Temp2
Calc_crc:
   Temp_crc = Temp2 * 256
   Crc = Temp_crc Xor Crc
   For Temp1 = 0 To 7
       If Crc > 32767 Then
          Crc = Crc * 2
          Crc = Crc Xor Crcpoly
       Else
          Crc = Crc * 2
       End If
       Crc = Crc And 65535
   Next
Return



' Send NAK (i.e tell master that packet was bad)
_nak:
   Hdb2 = Nakhdb2                                             ' Set ACK bits to NAK in HDB2 (xxxxxx11)
   Goto _ret
' Send Ack (i.e Tell Master That Packet was Ok)
_ack:
   Hdb2 = Ackhdb2                                             ' Set ACK bits to ACK in HDB2 (xxxxxx10)
_ret:
   Sab1 = Myaddress
   Dab1 = Hisaddress
   Sdb(1) = 0                                                 ' return for ACK / NAK has no value in db1
   Crc = 0                                                    ' Clear CRC variable
   Temp2 = Hdb2
   Gosub Calc_crc
   Temp2 = Ackhdb1
   Gosub Calc_crc
   Temp2 = Dab1
   Gosub Calc_crc
   Temp2 = Sab1
   Gosub Calc_crc
   Temp2 = Sdb(1)
   Gosub Calc_crc
   Crc2 = High(crc)
   Crc1 = Low(crc)
   Print Chr(preamble) ; Chr(sync) ; Chr(hdb2) ; Chr(ackhdb1) ; Chr(dab1) ; Chr(sab1) ; Chr(sdb(1)) ; Chr(crc2) ; Chr(crc1);
Return


' Subroutine for sending values
_send:
   Dab1 = Hisaddress                                          ' Change SAB1 and DAB1 address bytes
   Crc = 0                                                    ' Clear CRC variable
   Temp2 = Anshdb2                                            ' Put HDB2 in tempory variable. Always answering with defined format
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Chdb1r                                             ' Put HDB1 in tempory variable. Always answering with read flag
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Dab1                                               ' Put DAB1 in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Sab1                                               ' Put SAB1 in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Sdb(5)                                             ' Put DATA Byte 1 in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Sdb(4)                                             ' Put DATA Byte 2 in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Sdb(3)                                             ' Put DATA Byte 3 in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Sdb(2)                                             ' Put DATA Byte 4 in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC
   Temp2 = Sdb(1)                                             ' Put Command Byte in tempory variable
   Gosub Calc_crc                                             ' Calculate CRC

   Crc2 = High(crc)                                           ' Put calculated Hi_CRC value in CRC2
   Crc1 = Low(crc)                                            ' Put calculated Lo_CRC value in CRC1

   ' Send packet to master, including the SYNC byte
'      Print Chr(preamble) ; Chr(sync) ; Chr(anshdb2) ; Chr(chdb1r) ; Chr(dab1) ; Chr(sab1) ; Chr(sdb(5)) ; Chr(sdb(4)) ; Chr(sdb(3)) ; Chr(sdb(2)) ; Chr(sdb(1)) ; Chr(crc2) ; Chr(crc1)
   Print Chr(preamble) ; Chr(sync) ; Chr(anshdb2) ; Chr(&B00000101) ; Chr(dab1) ; Chr(sab1) ; Chr(sdb(5)) ; Chr(sdb(4)) ; Chr(sdb(3)) ; Chr(sdb(2)) ; Chr(sdb(1)) ;
Return


' -----[ Interrupt routines ]------------------------------------------------
'
' Interrupt routine for reading serial input
Rec_isr:
   If Bitcounter < Cmaxchar Then                              ' Does it fit into the buffer?
   Incr Bitcounter                                            ' Increase buffer counter
   Else
      Bitcounter = 1                                          ' Safety rewrap character serial counter
   End If
   Bitbuffer(bitcounter) = Udr                                ' Add to buffer array
   Incr Bitsreceived                                          ' Number of character flag
Return


' Interrupt routine for Stepper cycles
Timer0_isr:

Return

' Interrupt routine for PWM
Timer1_isr:
   Timer1_reload2 = Datas(13) * 256
   Timer1_reload2 = Timer1_reload2 + Datas(14)
   If Timer1_reload2 > &HFF Then
      Timer1 = Timer1_reload2
   Else
      Timer1 = Timer1_reload
   End If
   Db (1) = Db(1) - 1
   Sdb(1) = Db (1)                                            ' to avoid interuppt conflicts
   Sdb(2) = Datas(13)                                         ' to avoid interuppt conflicts
   Sdb(3) = Datas(14)                                         ' to avoid interuppt conflicts
   Sdb(4) = Datas(15)                                         ' to avoid interuppt conflicts
   Sdb(5) = Datas(16)                                         ' to avoid interuppt conflicts

   Gosub _send
Return